// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  AngleFormatter.cc                           XMM Science Analysis Subsystem
 *
 *  Angle with methods for parsing and formatting
 *
 *  (c) 1999  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: AngleFormatter.cc,v 1.8 2002/12/17 17:53:47 jbakker Exp $
 *
 *****************************************************************************/

#include <cmath>                    // C headers
#include <cstdio>
#include <iomanip>

#include "AngleFormatter.h"          // C++ headers
#include "StringFns.h"
#include <errstr.h>


//================================================================
//  Class AngleFormatter
//================================================================

/*** Construct a new angle (zero by default) ***/

AngleFormatter::AngleFormatter()
{
    _mode = Degrees;
}


AngleFormatter::AngleFormatter(AngleMode mode)
{
    _mode = mode;
}


/** Set the formatting mode ***/

void AngleFormatter::setMode(AngleMode mode)
{
    _mode = mode;
}


/*
 *  Format the angle as a std::string. The format is determined by the 'mode'
 *  (HMS/DMS/Degrees). e.g:
 *
 *      12h34m56.789s    (HMS)
 *      12d34m56.789s    (DMS)
 *      123.456          (Degrees)
 *
 *  If the 'separator' character is specified, it is used to separate
 *  the DMS/HMS fields instead of using "dms" / "hms" as delimiters.
 *
 *  Example:
 *      setMode(AngleFormatter::DMS);
 *      asString(12.3456, '#')  ->  12#20#44.16
 */

std::string AngleFormatter::asString(double value, char separator) const
{
    std::string result;

    switch(_mode)
    {
        case Degrees: {
	    // Decimal number with up to 6 decimal places 
  	    char buff[30];
	    sprintf(buff,"%.6f",value);

	    // The following is better, but not supported by DEC_OSF:
  	    // snprintf(buff,30,"%.6f",value);

	    // Remove trailing zeroes and trailing decimal point
	    std::string s(buff);
	    
	    std::string::size_type i = s.find_last_not_of('0');
	    if(i == std::string::npos)
	        result = "0";
	    else
	        if(s[i] == '.')
		    result = s.substr(0,i);
		else
		    result = s.substr(0,i+1);
	    break;
	}

        case DMS:
        case HMS: {
	    char delim1 = 'd';
	    if(_mode == HMS)
	        delim1 = 'h';
	    char delim2 = 'm';
	    char delim3 = 's';

	    if(separator != '\0')
	    {
	        delim1 = separator;
		delim2 = separator;
		delim3 = '\0';
	    }

	    /*
	     *  Notes:
	     *    (a)  -0:1:2 requires minus sign although first field is zero
	     *    (b)  1:2:59.9999 must not round to 1:2:60
	     *    (c)  Trailing zeroes are omitted from fractional degrees
	     */

	    double val = value;

	    if(_mode == HMS)
	        val /= 15.0;    // Convert arcseconds to seconds

	    // Round to nearest millisecond
	    long mis = (long)floor(3600000.0 * val + 0.5);
	    bool neg = (mis < 0);
	    long amis = labs(mis);

	    int f  = amis % 1000;
	    int tf = amis / 1000;
	    int s  = tf % 60;
	    int ts = tf / 60;
	    int m  = ts % 60;
	    int tm = ts / 60;

	    int x = (_mode == HMS) ? 24 : 360;
	    int h  = tm % x;

	    std::ostringstream ostr;
	    if(neg)
	        ostr << '-';
	    ostr << h << delim1 << m << delim2 << s;
	    if(f > 0)
	    {
	        ostr << "." << f / 100;
		int b = f % 100;
		if(b > 0)
		{
		    ostr << b / 10;
		    int d = b % 10;
		    if(d > 0)
		        ostr << d;
		}
	    }
	    ostr << delim3;
	    result = ostr.str();
	    break;
	}
    }
    return result;
}


/*
 *  Parse 'text' as an angle. If the angle is valid, return true, set
 *  the value and set the mode to the angle format (HMS/DMS/Degrees).
 *
 *  A leading minus sign applies to the entire angle, not just to the
 *  first field. (e.g: "-1d30m0s" = -1.5 degrees).
 *
 *  Examples of valid angle formats:
 *     12h34m56.789s    (HMS)
 *     12H34M56.789S
 *     12d34m56.789s    (DMS)
 *    -12d34m56.789s
 *     123.456          (Degrees)
 *    -123.456
 *
 *  If useColons is set, colons are accepted as separators, instead
 *  of the "dms" / "hms" delimiters. In this case the current mode
 *  is used to distinguish between DMS and HMS.
 *
 *  Example:
 *    12:34:56.789     (= 12h34m56.789s in HMS mode)
 *    12:34:56.789     (= 12d34m56.789s in DMS mode)
 */

bool AngleFormatter::parse(const std::string& str, double& result, char separator)
{
    double value = 0.0;
    bool ok = false;
    AngleMode mode = Degrees;

    std::string::size_type delim = str.find_first_not_of("0123456789+- ");
    if(delim != std::string::npos)
    {
        if(str[delim] == separator && separator != '\0')
	    ok = parseFields(str, separator, value, _mode);

	else switch(str[delim])
	{
	case 'h':
	    ok = parseFields(str,'h','m','s', value, HMS);
	    mode = HMS;
	    break;

	case 'H':
	    ok = parseFields(str,'H','M','S', value, HMS);
	    mode = HMS;
	    break;

	case 'd':
	    ok = parseFields(str,'d','m','s', value, DMS);
	    mode = DMS;
	    break;

	case 'D':
	    ok = parseFields(str,'D','M','S', value, DMS);
	    mode = DMS;
	    break;

	case '.':
	case 'e':
	case 'E':
	    mode = Degrees;
	    ok = StringFns::parseDouble(str, value);
	    break;

	default:
	    break;
	}

    } else {
	mode = Degrees;
	ok = StringFns::parseDouble(str, value);
    }

    if(value == -0.0)
        value = 0.0;

    if(ok)
    {
        result = value;
	_mode  = mode;
    }
    else {
       errstr << "Angle expected at '" << str << "'" << std::endl;
    }
    return ok;
}


/*
 *  Parse 'text' (in DMS or HMS format) as an angle, using delim1, delim2 
 *  and delim3 as field delimiters, and return its value via argument
 *  'result'. Return true if the angle is valid.
 *
 *  Example:
 *    ok = parseFields("12h34m56.789s",'h','m','s',result,HMS);
 */

bool AngleFormatter::parseFields(const std::string& text, char delim1, char delim2,
			 char delim3, double& result, AngleMode mode)
{
    bool ok = false;
    double value = 0.0;
    
    std::string::size_type i = text.find(delim1);
    if(i != std::string::npos)
    {
        std::string::size_type j = text.find(delim2,i+1);
	if(j != std::string::npos)
	{
	    std::string::size_type k = text.find(delim3,j+1);

	    if(k != std::string::npos)
	    {
	        int hrsDeg  = 0;
		int mins    = 0;
		double secs = 0.0;

		ok  = (text.length() == k+1);
		ok &= StringFns::parseInt(text.substr(0,i), hrsDeg);
		ok &= StringFns::parseInt(text.substr(i+1,j-i-1), mins);
		ok &= StringFns::parseDouble(text.substr(j+1,k-j-1), secs);
		ok &= convertFields(hrsDeg, mins, secs, (mode==HMS), value);
		// -0!!
		if (ok && value != 0. && hrsDeg==0) {
		  if (text.substr(0,i).find('-') != std::string::npos) value=-value;
		}
	    }
	}
    }
    if(ok)
        result = value;

    return ok;
}


/*
 *  Parse 'text' (in DMS or HMS format) as an angle, using specified
 *  'separator' character between fields, and return its value via 
 *  argument 'result'. Return true if the angle is valid.
 *
 *  Example:
 *    ok = parseFields("12:34:56.789",':',result,HMS);
 */

bool AngleFormatter::parseFields(const std::string& text, char separator,
				 double& result, AngleMode mode)
{
    bool ok = false;
    double value = 0.0;
    
    std::string::size_type i = text.find(separator);
    if(i != std::string::npos)
    {
        std::string::size_type j = text.find(separator,i+1);
	if(j != std::string::npos)
	{
	    std::string::size_type k = text.length();

	    int hrsDeg  = 0;
	    int mins    = 0;
	    double secs = 0.0;
	    
	    ok  = StringFns::parseInt(text.substr(0,i), hrsDeg);
	    ok &= StringFns::parseInt(text.substr(i+1,j-i-1), mins);
	    ok &= StringFns::parseDouble(text.substr(j+1,k-j-1), secs);
	    ok &= convertFields(hrsDeg, mins, secs, (mode==HMS), value);

	    // -0!!
	    if (ok && value != 0. && hrsDeg==0) {
	      if (text.substr(0,i).find('-') != std::string::npos) value=-value;
	    }

	}
    }
    if(ok)
        result = value;

    return ok;
}


/*
 *  Convert DMS/HMS fields into an angle. Return true if the fields are
 *  valid.
 *
 *  Example:
 *    ok = convertFields(12,34,56.789,true,result);
 */

bool AngleFormatter::convertFields(int hrsDeg,      // Hours or degrees field
			   int mins,        // Minutes fields
			   double secs,     // Seconds field
			   bool hoursMode,  // True for hours mode
			   double& result)  // Result in degrees
{
    bool ok = (-359 <= hrsDeg) && (hrsDeg <= 359)
      && (0 <= mins) && (mins <= 59)
      && (0.0 <= secs) && (secs < 60.0);

    if(hoursMode)
        ok &= (-23 <= hrsDeg) && (hrsDeg <= 23);
    
    if(ok)
    {
        double value = (double)abs(hrsDeg) + mins/60.0 + secs/3600.0;

	if(hrsDeg < 0)
	    value = -value;
		    
	if(hoursMode)
	    value *= 15.0;

        result = value;
    }
    return ok;
}


