// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include <sstream>             // C++ headers
#include <iostream>

#include "IntParameter.h"          // Application headers
#include "StringToken.h"
#include <errstr.h>
#include <Evaluator.h>

std::string IntParameter::asString() const
{
    std::ostringstream ostr;
    ostr << _value;
    return ostr.str();
}

std::string IntParameter::defaultAsString() const
{
    std::ostringstream ostr;
    ostr << _default;
    return ostr.str();
}

bool IntParameter::parse(const std::string& str, int& value)
{
    StringToken t(str);
    int n = t.asInt();
    if(t.isValid())
        value = n;
    else
      errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");

    return true;
}

void IntParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}

bool IntParameter::isValid(int n) const
{
  Evaluator evaluator( constraints() );
  NumericSymbols ids = evaluator.numericIdentifiers();
  if (ids.valuePtr.size() == 0) return evaluator.evalLogic();

  for (unsigned int i=0;i<ids.valuePtr.size();i++) *ids.valuePtr[i]=n;
  return evaluator.evalLogic();
}


bool IntParameter::setValue(int value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }
    return valid;
}


bool IntParameter::setDefault(int value)
{
    bool valid = isValid(value);
    if(valid)
        _default = value;
    return valid;
}


bool IntParameter::setNewValue(const std::string& str)
{
    int value;
    bool ok = parse(str, value) && setValue(value);

    if(!ok)
        reportRangeError(str);
    return true;
}


bool IntParameter::setNewDefault(const std::string& str)
{
    int value;
    bool ok = parse(str, value) && setDefault(value);

    if(!ok)
        reportRangeError(str);

    return true;
}


void IntParameter::reportRangeError(const std::string& valueString)
{
  errstr << "Parameter '" << _name << "';" << std::endl
	 << "value '" << valueString << "' is outside range ";
  writeConstraint(errstr);
  errstr << Sas::error("ParamNotInRange");
}
