// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// -*-C++-*-
//
// File:      XpiParamParser.cc
// Author:    Jorgo Bakker (jbakker@rssd.esa.int)
// Generated: Mon Feb 10 12:48:04 CET 2003
// Usage:     -
// Info:      -

#include "XpiParamParser.h"
#include "ParameterInfo.h"
#include "Parameter.h"      
#include "BoolParameter.h"      
#include "IntParameter.h"      
#include "RealParameter.h" 
#include "AngleParameter.h"
#include "TimeParameter.h"     
#include "StringParameter.h"
#include "SpecialStringParameters.h"
#include "StringParameter.h"
#include "BoolListParameter.h"
#include "IntListParameter.h"
#include "RealListParameter.h"
#include "AngleListParameter.h"
#include "TimeListParameter.h"
#include "StringListParameter.h"
#include "SpecialStringListParameters.h"
#include "StringToken.h"
#include <errstr.h>
#include <ExceptionErrorHandler.h>
#include <StringFns.h>

//#include <values.h>            // C headers
#include <climits>

#include <iostream>          // C++ headers
#include <fstream>
#include <string>
#include <sstream>
#include <list>
#include <vector>
#include <fstream>

#include <qdom.h>
#include <qfile.h>

/*---                    Forward declaration of the reader                ---*/
class XpiReader
{
public:
    XpiReader(const std::string& message) : _err(false), _message(message) {}
    ParameterInfo* parse(std::istream& is);
    bool isValid() const { return !_err; }

private:
    StringToken  getToken();

    std::istringstream*    _ss;
    ParameterInfo*         _info;
    bool                   _err;
    std::string            _message;
};

/*---                    Forward declaration of the writer                ---*/
class XpiWriter
{
public:
  char toXpiType(Parameter::ParamType t);
  std::string toMinMax(Parameter* p);
  void toXpi(std::ostream& os,Parameter* p);
};


/**
 */
XpiParamParser::XpiParamParser() 
{
}

XpiParamParser::~XpiParamParser()
{
}

ParameterInfo* XpiParamParser::load(const std::string& nameOfFile)
{

  errstr << "Using parameter file " <<  nameOfFile
	 << Sas::message(Msg::Lib, Msg::Noisy);
  ParameterInfo* result=0;
  
  std::ifstream is(nameOfFile.c_str());
  if (is) {
      XpiReader p(nameOfFile);
      result=p.parse(is);
  } else {
    errstr << "Could not open parameter file '"
	   << nameOfFile  << "'"
	   << Sas::error("ParamFileOpen");
  }
  return result;
}



void XpiParamParser::save(const ParameterInfo& info,const std::string& nameOfFile)
{
  // create output stream
  std::ofstream os(nameOfFile.c_str());

  // create header
  os << "# saved settings for task \""
     << info.taskName() << '"'
     << std::endl;

  // save all parameters
  XpiWriter w;
  std::list<std::string> parameters=info.names();
  for (std::list<std::string>::iterator it=parameters.begin();
       it != parameters.end();it++){
    Parameter* p=info[*it];
    if (!p) errstr << "Parameter[" << (*it) << "] not accessible"
		   << Sas::fatal("MemoryCorrupted");
    w.toXpi(os,p);
  }

  // create footer
  os << "# end"
     << std::endl;  
}



/*---                  Implementation of internal reader                  ---*/

/*** Get next token as a std::string ***/

StringToken XpiReader::getToken()
{
    const char quote ='"';

    std::string tok;
    char c = _ss->get();
    std::string token = "";

    // Skip leading space
    while(c == ' ' || c == '\t')
        c = _ss->get();

    while(c != EOF && c != ',' && c != '#') 
    {
        // Quotes - Copy everything up to closing quote
        if(c == quote) 
	{
	    token += quote;
	    c = _ss->get();
	    while(c != EOF && (c!=quote)) {
	        token += c;
		c = _ss->get();
	    }
	    if(c == EOF)
	      errstr << "Mismatched quotes reading token(" << token << ") in parameter file: "<< _message << Sas::error("ParamWrongQuotes");
	    token += quote;
	    c = _ss->get();
	    
	// Just copy anything else
	} else {
	    token += c;
	    c = _ss->get();
	}
    }

    // Remove trailing spaces
    // ("find_last_not_of" does not work with gcc-2.7.2.3)

    std::string::size_type n = token.length();
    while(n-- > 0 && (token[n] == ' ' || token[n] == '\t')) ;
    tok = token.substr(0,n+1);
    
    // Stop if comment found
    if(c == '#')
        _ss->putback(EOF);

    return StringToken(tok);
}


/*** Return a new ParameterInfo; caller must delete ***/

ParameterInfo* XpiReader::parse(std::istream& is)
{
  ExceptionErrorHandler local;
  local.name(errHandler.name());
  std::string buff;
  _info = new ParameterInfo;
  int lineno = 0;
  while(is) {
    
    getline(is,buff);
    lineno++;
    _ss = new std::istringstream(buff);
    
    Parameter* p=0;
    ErrorHandler* oldHandler = errHandler.handler(&local);
    std::string name = getToken().asUnquotedString();
    try {
      if(name != "") {
	std::string type      = getToken().asString();
	std::string mode      = getToken().asUnquotedString();
	StringToken def  = getToken();
	StringToken min  = getToken();
	StringToken max  = getToken();
	std::string prompt    = getToken().asUnquotedString();
	
	// mode = (|h|q)
	if ( !mode.empty() && !(mode == "h" || mode == "q") ) 
	  Sas::error("unrecognised mode \"" + mode + "\"\n");

	/* 
	   if max.empty() and not min.empty() and t='s|S'
	   then convert to items
	*/
	/* 
	   if t='b' then no constraints!
	*/
	std::ostringstream constraints;
	//string constraints("true");
	std::list<std::string> choices;
	
	
	/* 
	   TODO: Parameter class should handle errors!
	*/

	bool defaultUnquoted=false;
	char t = (type.length() > 0) ? type[0] : ' ';
	//try {
	switch (t) {
	case 'b':
	  p = new BoolParameter();
	  choices.push_back("yes");
	  choices.push_back("no");
	  break;
	  
	case 'i': p = new IntParameter();          // fall through!
	case 'r': if (!p) p = new RealParameter(); // fall through!
	case 'u': if (!p) p = new AngleParameter();// fall through!
	case 'w': if (!p) p = new TimeParameter();
	  defaultUnquoted=(t == 'w');
	  // all these may have constraints defined...we accept the following form
	  // [:double] | [double:] | [double:double]
	  if ( !(min.isUndef() && max.isUndef()) )
	    constraints << name << " in [" << min.asString()
			<< ":" << max.asString()<<"]";
	  break;
	  
	case 's': {
	  p = new StringParameter();
	  defaultUnquoted=true;
	  std::string minVal = min.asUnquotedString();
	  //if (minVal != "") ???
	  choices=StringFns::splitDelimitedString(minVal,"|");
	  break;
	}
	case 'y':
	  p = new DirectoryParameter();
	  defaultUnquoted=true;
	  break;
	case 'e':
	  p = new DataSetParameter();
	  defaultUnquoted=true;
	  break;
	case 'z':
	  p = new DirOrDataSetParameter();
	  defaultUnquoted=true;
	  break;
	case 'x':
	  p = new DirOrFileParameter();
	  defaultUnquoted=true;
	  break;
	case 'f':
	  p = new FileNameParameter();
	  defaultUnquoted=true;
	  break;
	case 'k':
	  p = new BlockParameter();
	  defaultUnquoted=true;
	  break;
	case 't':
	  p = new TableParameter();
	  defaultUnquoted=true;
	  break;
	case 'a':
	  p = new ArrayParameter();
	  defaultUnquoted=true;
	  break;
	case 'c':
	  p = new ColumnParameter();
	  defaultUnquoted=true;
	  break;
	case 'p':
	  p = new PlaneParameter();
	  defaultUnquoted=true;
	  break;
	  
	case 'B':
	  p = new BoolListParameter();
	  break;
	  
	case 'I': p = new IntListParameter();          // fall through!
	case 'R': if (!p) p = new RealListParameter(); // fall through!
	case 'U': if (!p) p = new AngleListParameter();// fall through!
	case 'W': if (!p) p = new TimeListParameter();
	  // all these may have constraints defined...we accept the following form
	  // [:double] | [double:] | [double:double]
	  if ( !(min.isUndef() && max.isUndef()) )
	    constraints << name << " in [" << min.asString()
			<< ":" << max.asString()<<"]";
	  break;
	  
	case 'S': {
	  p = new StringListParameter();
	  std::string minVal = min.asUnquotedString();
	  //if (minVal != "") ???
	  choices=StringFns::splitDelimitedString(minVal,"|");
	  break;
	}
	case 'Y':
	  p = new DirectoryListParameter();
	  break;
	case 'E':
	  p = new DataSetListParameter();
	  break;
	case 'F':
	  p = new FileNameListParameter();
	  break;
	case 'K':
	  p = new BlockListParameter();
	  break;
	case 'T':
	  p = new TableListParameter();
	  break;
	case 'A':
	  p = new ArrayListParameter();
	  break;
	case 'C':
	  p = new ColumnListParameter();
	  break;
	case 'P':
	  p = new PlaneListParameter();
	  break;
	default:
	  errstr << "Unrecognised parameter type:" << buff
		 << Sas::error("invalidParam");
	  break;
	}
	
	if (p) {
	  std::string theDefault=defaultUnquoted?def.asUnquotedString():def.asString();
	  std::string constraint=constraints.str();;
	  if (constraint.empty()) constraint="true";
	  bool isMandatory=( mode.find("q") != std::string::npos );

	  // now feed the information into the parameter...
	  p->setName(name);
	  p->setMandatory(isMandatory);
	  p->setPrompt(prompt);
	  p->setConstraints(constraint);
	  if (!theDefault.empty()) 
	    p->setNewDefault(theDefault);
	  
	  for(std::list<std::string>::const_iterator it=choices.begin();it!=choices.end();it++)
	    p->addChoice( *it );
	  
	  p->revertToDefault();

	  _info->addParameter(p);
	  p=0;
	}
      }
    }
    catch (SasException::Exception e) {


      errHandler.handler(oldHandler);
      // todo: delete p?
      _err = true;
      errstr << _message << ":" << lineno << ": " << e.message
	     << Sas::warning("invalidParam");
    }

    errHandler.handler(oldHandler);
  }
  delete _ss;
  _ss = 0;
    
  if (_err) {
    delete _info;
    _info=0;
  }
  return _info;
}

/*---                  Implementation of internal writer                  ---*/
char XpiWriter::toXpiType(Parameter::ParamType t)
{
  char c=0;
  switch (t) {
  case Parameter::Bool: 
    c='b';break;
  case Parameter::Int:
    c='i';break;
  case Parameter::Real:
    c='r';break;
  case Parameter::String:
    c='s';break;
  case Parameter::StringChoice: // same as 'String'?
    c='s';break;
  case Parameter::FileName:
    c='f';break;

  case Parameter::BoolList:
    c='B';break;
  case Parameter::IntList:
    c='I';break;
  case Parameter::RealList:
    c='R';break;
  case Parameter::StringList:
    c='S';break;
 
  case Parameter::StringChoiceList: // same as 'StringList'?
    c='S';break;
  case Parameter::FileNameList:
    c='F';break;

  case Parameter::Directory:
    c='y';break;
  case Parameter::DataSet:
    c='e';break;
  case Parameter::Block:
    c='k';break;
  case Parameter::Table:
    c='t';break;
  case Parameter::Array:
    c='a';break;
  case Parameter::Column:
    c='c';break;
  case Parameter::Plane:
    c='p';break;

  case Parameter::DirectoryList:
    c='Y';break;
  case Parameter::DataSetList:
    c='E';break;
  case Parameter::BlockList:
    c='K';break;
  case Parameter::TableList:
    c='T';break;

  case Parameter::ArrayList:
    c='A';break;
  case Parameter::ColumnList:
    c='C';break;
  case Parameter::PlaneList:
    c='P';break;

  case Parameter::Angle:
    c='u';break;
  case Parameter::Time:
    c='w';break;
  case Parameter::AngleList:
    c='U';break;
  case Parameter::TimeList:
    c='W';break;

  case Parameter::DirOrDataSet:
    c='z';break;
  case Parameter::DirOrFile:
    c='x';break; 
  case Parameter::Undefined:
    c=0;
  }

  return c;
}

std::string XpiWriter::toMinMax(Parameter* p)
{
  std::string r; // result;


  // option: r="a|b|c,"
  if (p->getType() == Parameter::Bool) {
    r=','; // empty!
    return r;
  }

  {
    std::list<std::string> ch=p->choices();
    for(std::list<std::string>::const_iterator it=ch.begin();it!=ch.end();it++) {
      if (it !=ch.begin()) r+='|';
      r+=(*it);
    }
    if (!ch.empty()) r+=','; // add the max part
  }
  if (!r.empty()) return r;


  r=',';

  // option: r=","
  const std::string& c=p->constraints();
  if (c.empty() || c == "true") {
    return r;
  }


  // option r="min,max"
  //    name in [low:high]
  //    name in (low,high)
  //    name in :high
  
  r=',';
  std::string::size_type idx=c.find(p->name());
  if (idx==std::string::npos) {
    errstr << "in " << p->name() << ", cannot parse constraints \""
	   << c << "\". Empty constraints created."
	   << Sas::warning("Xml2XpiConversion");
    return r;
  }
  idx+=p->name().length();

  idx=c.find(" in ",idx);
  if (idx==std::string::npos) {
    errstr << "in " << p->name() << ", cannot parse constraints \""
	   << c << "\". Empty constraints created."
	   << Sas::warning("Xml2XpiConversion");
  }
  idx+=4; // size of " in "

  // find the expression in between possible delimiters
  std::string::size_type be=c.find_first_of("[(",idx);
  if (be==std::string::npos) be=idx;
  else be+=1;

  std::string::size_type en=c.find_first_of(")]",be);
  if (en==std::string::npos) en=c.length();
  en-=1;
  if (en <= be) {
    return r;
  }

  // gather the min max
  r="";
  {
    std::list<std::string> ch=StringFns::splitDelimitedString(c.substr(be,en-be+1),":");
    for(std::list<std::string>::const_iterator it=ch.begin();it!=ch.end();it++) {
      if (it !=ch.begin()) r+=',';
      r+=(*it);
    }
  }
  return r;
}

void XpiWriter::toXpi(std::ostream& os,Parameter* p)
{
  char type=toXpiType(p->getType());
  char mandatory=p->isMandatory()?'q':'h';
  std::string constraints=toMinMax(p);

  os << p->name() << ','
     << type << ','
     << mandatory << ','
     << p->defaultAsString() << ','
     << constraints << ','
     << '"' << p->prompt() << '"'
     << std::endl;
}

